// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of src
//
// src is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// src is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <iostream>
#include <map>

#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stdexcept>
#include <unistd.h>

#include "Options.h"
#include "config.h"

using namespace std;
int usage( );


// To enable map to compare the key.
struct ltpid_t {
  bool operator( )( const pid_t a, const pid_t b ) const {
    return a < b;
  }
};

//
// The key of the map is the child-pid number
// The value of the map is the cmd used to run the program/script
//
typedef map< pid_t, pair< int, string>, ltpid_t > mymap;

class MyException : public runtime_error {

public:

  MyException( pid_t p ) : runtime_error( "" ), pid( p ) { }

  pid_t pid;

};

int make_pipe( int* fildes ) {

  if ( 0 != pipe( fildes ) ) {
    perror( "broken pipe!" );
    return EXIT_FAILURE;
  }

  return EXIT_SUCCESS;

}

static void monitor_children( mymap &the_children ) {

  int status;

      
  while ( the_children.size() > 0 ) {

    const pid_t child_pid = waitpid( -1, &status, 0 );


    if ( WIFEXITED( status ) != 0 ) {
    
      if ( 0 != WEXITSTATUS( status ) ) {

	throw MyException( child_pid );
      }

    } else {

      cerr << child_pid << " was sent a signal " << WTERMSIG(status) << '\n';
      throw MyException( child_pid );

    }

    the_children.erase( child_pid );
  }

}

static pid_t fork_child( const char* cmd ) {

  pid_t child_pid;

  switch( (child_pid=fork( )) ) {

  case -1:
    cerr << "Unable to fork cmd: " << cmd << endl;
    exit( -1 );

  case 0:

    if ( -1 == execlp( "/bin/sh", "sh", "-c", cmd, (char*) 0 ) ) {
      cerr << "Could not execute command '" << cmd << "'\n";
      exit( 1 );
    }

  }

  return child_pid;

}

static pid_t fork_child( const char* cmd, int fd[ 2 ] ) {

  pid_t child_pid;

  pipe( fd );

  switch( (child_pid=fork( )) ) {

  case -1:
    cerr << "Unable to fork cmd: " << cmd << endl;
    exit( -1 );

  case 0:

    close( fd[ 0 ] );                   // close read end of pipe
    dup2( fd[ 1 ], STDOUT_FILENO );     // make 1 same as write-to end of pipe
    close( fd[ 1 ] );

    if ( -1 == execlp( "/bin/sh", "sh", "-c", cmd, (char*) 0 ) ) {
      cerr << "Could not excecute command '" << cmd << "'\n";
      exit( 1 );
    }

  }

  return child_pid;

}

int main( int argc, char* argv[] ) {

  mymap the_children;

  try {

    clo::parser		parser;
    parser.parse( argc, argv );
    const clo::options& cmdline_options = parser.get_options( );
    const std::vector<std::string>& nonoptions = parser.get_non_options();

    // purposely do nothing, since the user has not requested to do anything.
    if ( 0 == nonoptions.size( ) )
      return EXIT_SUCCESS;

    /*
    if ( "stdin" != cmdline_options.input ) {

      if ( 0 != access( cmdline_options.input.c_str( ), R_OK ) ) {

	cerr << "Unable to access file '" << cmdline_options.input << "'\n";

	return -1;

      }

    }
    */

    // one command? just run it as is
    if ( 1 == nonoptions.size( ) ) {

      return system( nonoptions[ 0 ].c_str( ) );

    } else {
	
      const int last_cmd( nonoptions.size( ) - 1 );
      pid_t child_pid;

      // first through penultimate command are handled the same:
      //   * create pipe to use between current command and next
      //   * fork the child, which will inherit the current STDIN and attach
      //     it's STDOUT to the write side of the pipe
      //   * redefine STDIN to the read side of the aforementioned
      //     pipe so that the next loop will pass it on to the child
      

      for ( int cmd_idx = 0; cmd_idx < last_cmd ; cmd_idx++ ) {

	int fd[ 2 ];

	const char* cmd_ptr = nonoptions[ cmd_idx ].c_str( );

	child_pid = fork_child( cmd_ptr, fd );

	// set the child's pid as the key and the command string as the value
	the_children[ child_pid ] = pair<int,string>(cmd_idx, nonoptions[ cmd_idx ]);

	//
	// deal with pipe between process cmd_idx & 
	//
	close( fd[ 1 ] );
	dup2( fd[ 0 ], STDIN_FILENO );
	close( fd[ 0 ] );

      }


      // last child is a straight fork; inherits current STDIN (which
      // is connected to the penultimate commands STDOUT) and 
      // STDOUT, which is the original STDOUT

      child_pid = fork_child( nonoptions[ last_cmd ].c_str( ) );

      // update the list of children

      the_children[ child_pid ] = pair<int,string>(last_cmd, nonoptions[ last_cmd ]);

      monitor_children( the_children );

      return EXIT_SUCCESS;

    }

  } catch ( clo::autoexcept & cloa ) {

    switch( cloa.get_autothrow_id( ) ) {
      
    case clo::autothrow_help:
      std::cout << "Usage: " << argv[0] << " [options]\n";
      std::cout << cloa.what();
      return EXIT_SUCCESS;

    case clo::autothrow_usage:
      usage( );
      return EXIT_SUCCESS;

    case clo::autothrow_version:
      std::cout << argv[ 0 ] << '\t' << VERSION << endl;
      return EXIT_SUCCESS;
    }

  } catch( MyException& me ) {

    mymap::const_iterator iterator = the_children.find( me.pid );
    cerr << argv[0] << ": Error executing cmd #" << iterator->second.first+1 << ":  "
	 << iterator->second.second << "\n";

    the_children.erase( me.pid );

    // get rid of the remaining children. we can't just kill( 0, SIGTERM), as
    // that does this process in too, which we don't want.  we want a
    // graceful exit

    for ( mymap::const_iterator child = the_children.begin() ;
	  child != the_children.end() ; child++ )
      kill( child->first, SIGTERM );

    // wait for them to all die
    while ( the_children.size() > 0 )
      the_children.erase( waitpid( -1, NULL, 0 ) );

    /*
    char mykillcmd[ 128 ];
    sprintf( mykillcmd, "kill %d", getpid( ) );
    execlp( "/bin/sh", "sh", "-c", mykillcmd, (char*) 0 );
    */

    return EXIT_FAILURE;

  } catch( std::exception &e ) {

    std::cerr << e.what() << "\n";
    return EXIT_FAILURE;

  } catch ( ... ) {

    cerr << "Unknown exception caught, got to love C++\n";
    return EXIT_FAILURE;

  }

}
